/* -*- mode: c; c-basic-offset: 4; -*- */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "daqc.h"
#include "daqc_response.h"
#include "daqc_internal.h"
#include "nds1.h"

/* #define DAQC_ACCESS_VERBOSE_ERRORS */

/*
 *  Connect to the DAQD server on the host identified by `ip' address.
 *  Returns zero if OK or the error code if failed.
 */
int
nds1_connect (daq_t *daq, const char* host, int port) {
    int resp;
    int v, rv;
    int on = 1;
    int connect_rc = 0;
    int cntr;

    /*  Get a socket.
     */
    if ((daq -> sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket() error");
	return DAQD_SOCKET;
    }

    setsockopt (daq->sockfd, SOL_SOCKET, SO_REUSEADDR, 
		(const char *) &on, (socklen_t) sizeof (on));

    /*----------------------------------  Get the server address.     */
    resp = daq_set_server(daq, host, port);
    if (resp) {
	close(daq->sockfd);
	daq->sockfd = -1;
	return resp;
    }

    /*-----------------------------------  Connect to the server      */
    fcntl(daq -> sockfd, F_SETFL, O_NONBLOCK);
    printf("Connecting..");
    fflush(stdout);
#define MAX_CONNECT_RETRY 4
    for (cntr=0; cntr<MAX_CONNECT_RETRY; ++cntr) {
	connect_rc = connect (daq->sockfd, 
			      (struct sockaddr *) &daq -> srvr_addr, 
			      (socklen_t) sizeof (daq -> srvr_addr)
			      );
	if (connect_rc < 0 && (errno == EINPROGRESS || errno == EALREADY)) {
	    sleep(1);
	    printf("..");
	    fflush(stdout);
	} else {
	    break;
	}
    }
    if (connect_rc < 0 && errno != EISCONN) {
	printf(" failed\n");
	perror("connect() error");
	close (daq -> sockfd);
	daq->sockfd = -1;
	return DAQD_CONNECT;
    }
    printf(" done\n");
    fcntl(daq -> sockfd, F_SETFL, 0);

    // Doing this does not work on Solaris
#ifdef __linux__
    {
	int fl = fcntl(daq -> sockfd, F_GETFL, 0);
	fcntl(daq -> sockfd, F_SETFL, O_NONBLOCK);

	/* Do protocol version check */
	if ((resp = daq_send (daq, "version;"))) {
	    close (daq -> sockfd);
	    daq->sockfd = -1;
	    return resp;
	}
	fcntl(daq -> sockfd, F_SETFL, fl);
    }
#else
    /* Do protocol version check */
    if ((resp = daq_send (daq, "version;"))) {
	close (daq -> sockfd);
	daq->sockfd = -1;
	return resp;
    }
#endif

    /* Read server response */
    if ((v = read_server_response(daq->sockfd)) != DAQD_PROTOCOL_VERSION) {
	fprintf(stderr, "communication protocol version mismatch: ");
	fprintf(stderr, "expected %d, received %d\n", DAQD_PROTOCOL_VERSION, v);
	close (daq -> sockfd);
	daq->sockfd = -1;
	return DAQD_VERSION_MISMATCH;
    }

    /* Do protocol revision check */
    resp = daq_send (daq, "revision;");
    if (resp) {
	close (daq -> sockfd);
	daq->sockfd = -1;
	return resp;
    }

    /* Read server response */
    rv = read_server_response(daq->sockfd);
    if (rv != DAQD_PROTOCOL_REVISION) {
	fprintf(stderr, "Warning: communication protocol revision mismatch: ");
	fprintf(stderr, "expected %d.%d, received %d.%d\n",
		DAQD_PROTOCOL_VERSION, DAQD_PROTOCOL_REVISION, v, rv);
    }
    daq -> rev = rv;
    daq -> datafd = daq -> sockfd;

#ifdef VXWORKS
    {
	int blen = 4*4096;
	if (setsockopt (daq -> sockfd, SOL_SOCKET, SO_SNDBUF, &blen, 
			sizeof (blen)) < 0) {
	    fprintf (stderr, "setsockopt SO_SNDBUF failed\n");
	    close (daq -> sockfd);
	    return DAQD_SETSOCKOPT;
	}
    }
#endif
    return 0;
}


/*
 *  Disconnect from the server and close the socket file descriptor
 */
int
nds1_disconnect (daq_t *daq) {
    char *command = "quit;\n";
    size_t  lcmd = strlen(command);
    if (write (daq -> sockfd, command, lcmd) != lcmd) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf (stderr, "daq_disconnect: write errno=%d\n", errno);
#endif
	return DAQD_WRITE;
    }
    usleep(100000);
    close (daq -> sockfd);
    return 0;
}


static void
cvthex(char** s, char* buf, void* x) {
    memcpy(buf, *s, (size_t)8);
    buf[8] = 0;
    (*s) += 8;
    *(int*)x = dca_strtol(buf);
}

/*
 *  Receive channel data using the NDS1 protocol
 */
int
nds1_recv_channels (daq_t *daq, daq_channel_t *channel, int num_channels, 
		    int *num_channels_received) {
    int i;
    int resp;
    char buf[129];
    size_t lChanWd = 8;
    int    channels = -1;
    FILE* f;

    const char* chan_req = "status channels 2;\n";
    resp = daq_send(daq, chan_req);
    if (resp) {
	return resp;
    }

    resp = read_bytes(daq->sockfd, buf, lChanWd);
    if (resp == lChanWd) {
	buf[lChanWd] = 0;
	channels = dca_strtol(buf);
    }

    if (channels <= 0) {
	fprintf (stderr, "Couldn't determine the number of data channels\n");
	return DAQD_ERROR;
    }

    *num_channels_received = channels;

    /*  Open the socket as a file
     */
    f = fdopen(dup(daq -> sockfd), "r");

    for (i = 0; i < channels; i++) {
	if (fgets(buf, (int) sizeof(buf), f) <= 0) {
	    perror("fgets failed");
	    fclose(f);
	    return DAQD_ERROR;
	}
	/* printf("chan string: %s\n", buf); */

	if (i < num_channels) {
	    int intval;
	    char* s;

	    /* Channel name                                 */
	    strncpy(channel[i].name, buf, (size_t)40);
	    s = strchr(channel[i].name, ' ');
	    if (s) *s = 0;
	    s = buf + 40;

	    /* Read rate                                    */
	    cvthex(&s, buf, &intval);
	    channel[i].rate = intval;

	    /* Read testpoint number                        */
	    cvthex(&s, buf, &channel[i].tpnum);

	    /* Group number and data type are 16-bit fields */
	    cvthex(&s, buf, &intval);
	    channel[i].type = cvt_group_chantype(intval >> 16);
	    channel[i].data_type = (intval & 0xffff);
	    channel[i].bps= (int)data_type_size(channel[i].data_type);

	    /* Read gain,slope and offset                   */
	    cvthex(&s, buf, &channel[i].s.signal_gain);
	    cvthex(&s, buf, &channel[i].s.signal_slope);
	    cvthex(&s, buf, &channel[i].s.signal_offset);

	    /* Read units                                   */
	    strncpy(channel[i].s.signal_units, s, (size_t)40);

	    /* printf("channel: %s  rate: %f  dtype: %i ctype: %i\n",
		   channel[i].name, channel[i].rate, channel[i].data_type,
		   intval); */
	}
    }
    fclose(f);
    return 0;
}

/*  nds1_request_data(daq_t*, time_t, time_t)
 *
 *  Get requested channel data for the specified interval.
 */
int
nds1_request_data(daq_t* daq, time_t start, time_t dt) {
    enum chantype reqty   = cUnknown;
    size_t        namel   = 0;
    size_t        txt_len = 0;    
    uint4_type    i;
    char          *cmd_txt, *p, *cmd;
    int           rc =DAQD_OK;

    /*  Get the request type and sum channel name length
     */
    uint4_type N = daq->num_chan_request;
    if (N == 0) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "No channels requested\n");
#endif
	return DAQD_ERROR;
    }

    for (i=0; i<N; ++i) {
	if (reqty == cUnknown) {
	    reqty = daq->chan_req_list[i].type;
	} else if (reqty != daq->chan_req_list[i].type) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	    fprintf(stderr, "Incompatible channel types\n");
#endif
	    return DAQD_ERROR;
	}
	namel += strlen(daq->chan_req_list[i].name) + 1;
    }

    /*----------------------------------  Resolve  Unknown channel types
     */
    if (reqty == cUnknown) {
	if (!start) reqty = cOnline;
	else        reqty = cRaw;
    }

    /*----------------------------------  Build up a command
     */
    cmd = NULL;
    switch (reqty) {

    /*----------------------------------  Online channel type.
     */
    case cOnline:
    case cRaw:
	cmd = "start net-writer";
	break;

    /*----------------------------------  Second trend channel type.
     */
    case cSTrend:
	cmd = "start trend net-writer";
	break;

    /*----------------------------------  Minute trend channel type.
     */
    case cMTrend:
	cmd = "start trend 60 net-writer";
	break;

    /*----------------------------------  Channel type not supported by NDS1
     */
    default:
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "Channel type %s not supported\n",
		cvt_chantype_str(reqty));
#endif
	return DAQD_ERROR;
    }

    /*----------------------------------  Get the text length, allocat a buffer
     */
    txt_len = strlen(cmd)  /* length of commmand text */
	    + strlen(" xxxxxxxxxx xxxxxxxxxx {};") /* punctuation, times */
	    + namel + N; /* length of channel names, blanks and NULL*/
    cmd_txt = (char*)malloc(txt_len);

    /*----------------------------------  Build the command
     */
    p = cmd_txt;
    while (*cmd) *p++ = *cmd++;
    *p++ = ' ';
    p += _daq_cvt_string(start, p);
    *p++ = ' ';
    p += _daq_cvt_string(dt,    p);
    *p++ = ' '; *p++ = '{';
    for (i=0; i<N; i++) {
	if (i) *p++ = ' ';
	*p++ = '"';
	char* pchn = daq->chan_req_list[i].name;
	while (*pchn) *p++ = *pchn++;
	*p++ = '"';
	daq->chan_req_list[i].status = 0;
    }
    *p++ = '}'; *p++ = ';';  *p++ = 0;  

    /*----------------------------------  Send the command.
     */
    rc = daq_send(daq, cmd_txt);
    if (rc) {
	printf("Error in daq_send: %i\n", rc);
    } else {
	int wid = daq_recv_id(daq);
	if (wid < 0) {
	    printf("Error in daq_recv_id: %i\n", wid);
	} else {
	    int bl = daq_recv_block_num(daq);
	    if (bl<0) printf("Error in daq_recv_block_num = %i\n", bl);
	}
    }

    /*----------------------------------  Free the command.
     */
    free(cmd_txt);
    cmd_txt = 0;
    return rc;
}

/*  nds1_receive_reconfigure(daq_t* daq, long block_len)
 *
 *  Receive a reconfigure block. receive_reconfigure is received after the 
 *  block header has been read in. The block length does not include the
 *  header length.
 */
int 
nds1_receive_reconfigure(daq_t* daq, size_t block_len) {
    size_t nchannels, alloc_size;
    int fd, i;
    signal_conv_t1* ps;

    nchannels = block_len / sizeof(signal_conv_t1);
    if (nchannels * sizeof(signal_conv_t1) != block_len) {
	fprintf (stderr, 
		 "warning:channel reconfigure block length has bad length"
		 );
    }
    if (nchannels == 0) return -2;
    alloc_size = sizeof(signal_conv_t1)*nchannels;

    if (!daq->s) {
	daq->s = (signal_conv_t1 *) malloc (alloc_size);
	if (daq -> s) {
	    daq -> s_size = (int) nchannels;
	} else {
	    fprintf(stderr, "malloc(%zi) failed; errno=%d\n", 
		    alloc_size, errno);
	    daq -> s_size = 0;
	    return -1;
	}
    } else if (daq->s_size != nchannels) {
	daq->s = (signal_conv_t1 *) realloc((void *)daq->s, alloc_size);
	if (daq->s) {
	    daq->s_size = (int) nchannels;
	} else {
	    fprintf (stderr, "realloc(%zi) failed; errno=%d\n", 
		     alloc_size, errno);
	    daq->s_size = 0;
	    return -1;
	}
    }

    fd = daq->datafd;
    ps = daq->s;
    for (i = 0; i < nchannels; i++) {
	if (read_float(fd, &ps[i].signal_offset)) 
	    return DAQD_ERROR;

	if (read_float(fd, &ps[i].signal_slope))
	    return DAQD_ERROR;
	read_uint4(fd, &ps[i].signal_status);
    }
    return -2; 
}

/*
 *  Initialize the nds1 protocol.
 */
int
nds1_startup(void) {
    return 0;
}
